__all__ = ['constraint_example', 'custom_ec_example', 'dea_example', 'eda_example', 'es_example', 'ga_example',
           'infinite_monkey_theorem', 'niche_example', 'nsga_example', 'paes_example', 'pso_example', 'sa_example']